/*
 * Copyright (c) 2019, Oracle and/or its affiliates. All rights reserved.
 */
package com.oracle.iot.flir;

import java.io.Closeable;
import java.io.File;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.websocket.DeploymentException;

import com.oracle.iot.flir.model.FLIRCameraE95DeviceDescriptor;
import com.oracle.iot.flir.model.FLIRCameraE95DeviceModelDescriptor;
import com.oracle.iot.flir.utils.Configuration;
import com.oracle.iot.flir.utils.FLIRConnection;
import com.oracle.iot.flir.utils.FLIRConnection.TelemetryHandler;
import com.oracle.iot.flir.utils.FLIRConnection.TemperatureHandler;
import com.oracle.iot.flir.utils.ProvisioningUtils;
import com.oracle.iot.flir.utils.ProvisioningUtils.DCDDescriptor;
import com.oracle.iot.flir.utils.ProvisioningUtils.RuleIncident;
import com.oracle.iot.flir.utils.http.HttpSecureConnection;

import oracle.iot.client.device.DirectlyConnectedDevice;
import oracle.iot.client.device.VirtualDevice;

public class FLIRIntegration {
    private static final String CONFIG_FILE_NAME = "config.json";
    
    private static final String ASSET_TYPE_NAME = "FLIRCamera";
    
    private static final String ASSET_NAME = "FLIR E95";
    
    private static final String TEMPERATURE_TOO_HIGH_RULE_NAME = "Temperature Too High";

    private static final Logger LOGGER = Logger.getLogger(FLIRIntegration.class.getSimpleName());

    public static void main(String[] args) {
        Configuration config;
        FLIRConnection flirConnection;

        try {
            config = new Configuration(new File(CONFIG_FILE_NAME));
            flirConnection = new FLIRConnection(config);
        } catch (IOException ex) {
            LOGGER.severe(ex.getMessage());

            return;
        }

        final VirtualDevice flirCamera;

        try {
            String amAppID = ProvisioningUtils.getAMAppId(config.getIOTConnection());
            deprovision(amAppID, config);
            DirectlyConnectedDevice flirDCD = provision(amAppID, config);

            if (flirDCD == null) {
                return;
            }

            flirCamera = flirDCD.createVirtualDevice(flirDCD.getEndpointId(),
                    flirDCD.getDeviceModel(FLIRCameraE95DeviceModelDescriptor.DEVICE_MODEL_URN));
        } catch (IOException | GeneralSecurityException ex) {
            LOGGER.log(Level.SEVERE, "Cannot provision IOT instance", ex);
            
            return;
        }

        Map<String, TelemetryHandler> handlers = new HashMap<>();

        handlers.put(".image.sysimg.measureFuncs.spot.1.valueT",
                new TemperatureHandler(config.getCameraTemperatureUnits()) {
                    @Override
                    protected void handleTemperature(Double value) {
                        flirCamera
                            .update()
                            .set(FLIRCameraE95DeviceModelDescriptor.ATTRIBUTE_TEMPERATURE, value)
                            .finish();
                    }
                });

        try (Closeable session = flirConnection.startTelemetryNotificationSession(handlers)) {
            System.out.println("Press [Enter] to exit");

            System.in.read();
        } catch (IOException | DeploymentException ex) {
            LOGGER.log(Level.SEVERE, "Cannot connect to FLIR camera sensors", ex);

            return;
        }
    }

    private static void deprovision(String amAppID, Configuration config) throws IOException, GeneralSecurityException {
        HttpSecureConnection iotConnection = config.getIOTConnection();

        String assetID = ProvisioningUtils.getAssetID(iotConnection, ASSET_NAME);
        
        if (assetID != null) {
            ProvisioningUtils.deleteAsset(iotConnection, assetID);
        }
        
        String endpointID = ProvisioningUtils.getDeviceIDByHardwareID(iotConnection,
                new FLIRCameraE95DeviceDescriptor(config).getHardwareID());

        if (endpointID != null) {
            ProvisioningUtils.decommissionDevice(iotConnection, endpointID);
            ProvisioningUtils.deleteDevice(iotConnection, endpointID);
        }

        String incidentRuleID = ProvisioningUtils.getRuleID(iotConnection, TEMPERATURE_TOO_HIGH_RULE_NAME);
        
        if (incidentRuleID != null) {
            ProvisioningUtils.deleteRule(iotConnection, incidentRuleID);
        }

        
        String assetTypeID = ProvisioningUtils.getAssetTypeID(iotConnection, ASSET_TYPE_NAME);
        
        if (assetTypeID != null) {
            ProvisioningUtils.deleteAssetType(iotConnection, assetTypeID);
        }

        ProvisioningUtils.ensureDeviceModelsAreDeletedFromApplication(iotConnection, amAppID,
                FLIRCameraE95DeviceModelDescriptor.DEVICE_MODEL_URN);

        if (ProvisioningUtils.deviceModelExists(iotConnection, FLIRCameraE95DeviceModelDescriptor.DEVICE_MODEL_URN)) {
            ProvisioningUtils.deleteDeviceModel(iotConnection, FLIRCameraE95DeviceModelDescriptor.DEVICE_MODEL_URN);
        }
    }

    private static DirectlyConnectedDevice provision(String amAppID, Configuration config)
            throws IOException, GeneralSecurityException {
        FLIRCameraE95DeviceModelDescriptor deviceModelDescriptor = new FLIRCameraE95DeviceModelDescriptor();
        HttpSecureConnection iotConnection = config.getIOTConnection();

        if (!ProvisioningUtils.deviceModelExists(iotConnection, FLIRCameraE95DeviceModelDescriptor.DEVICE_MODEL_URN)) {
            ProvisioningUtils.createDeviceModel(iotConnection, deviceModelDescriptor);
        }

        ProvisioningUtils.ensureDeviceModelsAreAddedToApplication(iotConnection, amAppID,
                FLIRCameraE95DeviceModelDescriptor.DEVICE_MODEL_URN);

        if (ProvisioningUtils.getAssetTypeID(iotConnection, ASSET_TYPE_NAME) == null) {
            ProvisioningUtils.createAssetType(iotConnection, ASSET_TYPE_NAME, deviceModelDescriptor);
        }
        
        if (ProvisioningUtils.getRuleID(iotConnection, TEMPERATURE_TOO_HIGH_RULE_NAME) == null) {
            ProvisioningUtils.createTooHighTemperatureRule(config, TEMPERATURE_TOO_HIGH_RULE_NAME, ASSET_TYPE_NAME, 
                    new RuleIncident("Temperature exceeded maximum value", "MAINTENANCE", "HIGH"));
        }

        DCDDescriptor deviceDescriptor = new FLIRCameraE95DeviceDescriptor(config);

        if (ProvisioningUtils.getDeviceIDByHardwareID(config.getIOTConnection(),
                deviceDescriptor.getHardwareID()) == null) {
            ProvisioningUtils.registerDCD(iotConnection, deviceDescriptor);
        }

        DirectlyConnectedDevice result = ProvisioningUtils.getDirectlyConnectedDevice(deviceDescriptor,
                FLIRCameraE95DeviceModelDescriptor.DEVICE_MODEL_URN);
        
        if (ProvisioningUtils.getAssetID(iotConnection, ASSET_NAME) == null) {
            ProvisioningUtils.createAsset(iotConnection, ASSET_NAME, ASSET_TYPE_NAME, config, result.getEndpointId(),
                    deviceModelDescriptor);
        }
        
        return result;
    }
}
