/*
 * Copyright (c) 2019, Oracle and/or its affiliates. All rights reserved.
 */
package com.oracle.iot.flir.utils;

import java.io.Closeable;
import java.io.IOException;
import java.net.NetworkInterface;
import java.net.SocketException;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.websocket.*;

import org.glassfish.tyrus.client.ClientManager;
import org.glassfish.tyrus.container.grizzly.client.GrizzlyClientContainer;
import org.json.JSONObject;

import com.oracle.iot.flir.utils.NetworkUtils.IPv4Subnet;
import com.oracle.iot.flir.utils.http.HttpPlainConnection;
import com.oracle.iot.flir.utils.http.HttpResponse;

public class FLIRConnection {
    private static final Logger LOGGER = Logger.getLogger(FLIRConnection.class.getSimpleName());

    private Configuration config;

    private String address;

    private HttpPlainConnection connection;
    
    private Map<String, String> headers;

    public FLIRConnection(Configuration config) throws IOException {
        this.config = config;

        String address = findFLIRAddress();

        if (address == null) {
            throw new IOException("FLIR Camera is not found");
        }

        LOGGER.info("FLIR camera is found at address " + address);

        this.address = address;

        loginAndInitialize();
    }

    private void loginAndInitialize() throws IOException {
        LOGGER.info("Logging in to FLIR camera and initializing telemetry sensors");

        HttpResponse response;

        try {
            connection = new HttpPlainConnection("http://" + address, null, null);
            response = connection.post("check_login",
                    ("_username=" + config.getCameraUserName() + "&_password=" + config.getCameraPassword())
                            .getBytes(StandardCharsets.UTF_8),
                    "application/x-www-form-urlencoded");
        } catch (GeneralSecurityException | IOException ex) {
            LOGGER.log(Level.SEVERE, "Cannot connect to FLIR camera", ex);

            response = null;
        }

        if (response == null || response.getStatus() != 302) {
            throw new IOException("Cannot login to FLIR camera");
        }

        Map<String, List<String>> responseHeaders = response.getHeaders();

        this.headers = new HashMap<>();

        copyHeader(responseHeaders, "Set-Cookie", "Cookie");
        copyHeader(responseHeaders, "Cache-Control");
        copyHeader(responseHeaders, "Pragma");

        try {
            response = connection.get("", "text/html", this.headers);
        } catch (GeneralSecurityException ex) {
            LOGGER.log(Level.SEVERE, "Cannot initialize telemetry sensors", ex);

            response = null;
        }

        if (response == null || response.getStatus() != 200) {
            throw new IOException("Cannot initialize telemetry sensors");
        }
    }

    private void copyHeader(Map<String, List<String>> headers, String header) {
        copyHeader(headers, header, header);
    }

    private void copyHeader(Map<String, List<String>> headers, String srcHeader, String dstHeader) {
        if (headers.containsKey(srcHeader)) {
            this.headers.put(dstHeader, headers.get(srcHeader).get(0));
        }
    }

    private boolean belongsToFLIR(IPv4Subnet subnet) {
        return subnet != null && subnet.matchPattern(((192L << 8) + 168L) << 16);
    }

    private String findFLIRAddress() {
        Enumeration<NetworkInterface> nis;

        try {
            nis = NetworkInterface.getNetworkInterfaces();
        } catch (SocketException ex) {
            LOGGER.log(Level.SEVERE, "Cannot get network interfaces", ex);

            return null;
        }

        while (nis.hasMoreElements()) {
            NetworkInterface ni = nis.nextElement();

            IPv4Subnet subnet = NetworkUtils.getIPv4Subnet(ni);

            if (belongsToFLIR(subnet)) {
                return subnet.getAddress(1);
            }
        }

        return null;
    }

    public Closeable startTelemetryNotificationSession(Map<String, TelemetryHandler> handlers)
            throws DeploymentException, IOException {
        LOGGER.info("Starting FLIR camera telemetry notifications session");

        TelemetryEndpoint endpoint = new TelemetryEndpoint(handlers);

        ClientManager.createClient(GrizzlyClientContainer.class.getName()).connectToServer(endpoint,
                URI.create("ws://" + address + "/x"));

        return endpoint;
    }
    
    public byte[] getSnapshot() throws IOException, GeneralSecurityException {
        HttpResponse response = connection.get("snapshot.jpg", "image/jpeg", headers);
        
        if (response.getStatus() != 200) {
            throw new IOException("Cannot get camera snapshot");
        }
        
        return response.getData();
    }


    private static class TelemetryEndpoint extends Endpoint implements MessageHandler.Whole<String>, Closeable {
        private Map<String, TelemetryHandler> handlers;
        private Session session;

        private TelemetryEndpoint(Map<String, TelemetryHandler> handlers) {
            this.handlers = handlers;
        }

        @Override
        public void close() throws IOException {
            LOGGER.info("Closing FLIR camera telemetry notifications session");

            session.close();
        }

        @Override
        public void onOpen(Session session, EndpointConfig config) {
            LOGGER.info("Subscribing to FLIR camera telemetry notifications");
            session.addMessageHandler(this);

            this.session = session;
        }

        @Override
        public void onMessage(String message) {
            try {
                if (handlers != null) {
                    JSONObject json = new JSONObject(message);

                    for (String key : json.keySet()) {
                        TelemetryHandler handler = handlers.get(key);

                        if (handler != null) {
                            handler.handle(json.get(key));
                        }
                    }
                }
            } catch (Throwable t) {
                // ignore
            }
        }
    }


    public static interface TelemetryHandler {
        void handle(Object value);
    }


    public static abstract class TemperatureHandler implements TelemetryHandler {
        private TemperatureUnits units;

        public TemperatureHandler(TemperatureUnits units) {
            this.units = units;
        }

        @Override
        public final void handle(Object value) {
            Double temperature = (Double) value;

            switch (units) {
                case Celsius:
                    temperature = temperature - 273.15;
                    break;
                case Fahrenheit:
                    temperature = (temperature - 273.15) * 9.0 / 5.0 + 32;
                    break;
                default:
                    break;
            }

            handleTemperature(temperature);
        }

        protected abstract void handleTemperature(Double value);
    }


    public static enum TemperatureUnits {
        Celsius, Fahrenheit, Kelvin;
    }
}
