/*
 * Copyright (c) 2019, Oracle and/or its affiliates. All rights reserved.
 */
package com.oracle.iot.flir.utils;

import java.net.Inet4Address;
import java.net.InetAddress;
import java.net.InterfaceAddress;
import java.net.NetworkInterface;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

public class NetworkUtils {
    private static final Pattern IPV4_ADDRESS_PATTERN = Pattern
            .compile("(\\d{1,3})\\.(\\d{1,3})\\.(\\d{1,3})\\.(\\d{1,3})");

    private NetworkUtils() {}
    
    public static IPv4Subnet getIPv4Subnet(NetworkInterface ni) {
        IPv4Subnet result = new IPv4Subnet(ni);
        
        return result.netmask > 0 ? result : null;
    }
    
    public static boolean isIPv4Address(String str) {
        return IPV4_ADDRESS_PATTERN.matcher(str).matches();
    }
    
    public static class IPv4Subnet {
        private long networkAddress;
        private long broadcastAddress;
        private long netmask;

        private IPv4Subnet(NetworkInterface ni) {
            List<InterfaceAddress> iAddresses = ni.getInterfaceAddresses();

            for (InterfaceAddress iAddress : iAddresses) {
                InetAddress address = iAddress.getAddress();

                if (address instanceof Inet4Address && !address.isLoopbackAddress() && address.isSiteLocalAddress()) {
                    int bitsToClear = 32 - iAddress.getNetworkPrefixLength();
                    
                    this.netmask = (0xFFFFFFFFL >> bitsToClear) << bitsToClear;
                    this.networkAddress = (toLong(address.getHostAddress()) & this.netmask);
                    this.broadcastAddress = (this.networkAddress | ~(this.netmask)) & 0xFFFFFFFFL;

                    break;
                }
            }
        }

        private long toLong(String address) {
            Matcher matcher = IPV4_ADDRESS_PATTERN.matcher(address);

            matcher.matches();

            long result = 0;

            for (int i = 1; i <= 4; ++i) {
                int addressPart = Integer.parseInt(matcher.group(i));
                result = result | ((addressPart & 0xFF) << 8 * (4 - i));
            }

            return result;
        }

        private String toString(long longAddress) {
            int parts[] = new int[4];

            for (int i = 3; i >= 0; --i) {
                parts[i] = (int) ((longAddress >>> 8 * (3 - i)) & (0xFF));
            }

            StringBuilder result = new StringBuilder();
            for (int i = 0; i < parts.length; ++i) {
                result.append(parts[i]);
                if (i != parts.length - 1) {
                    result.append(".");
                }
            }
            return result.toString();
        }

        private long startAddress() {
            return broadcastAddress - networkAddress > 1 ? networkAddress + 1 : 0;
        }

        private long endAddress() {
            return broadcastAddress - networkAddress > 1 ? broadcastAddress - 1 : 0;
        }

        public String getBroadcastAddress() {
            return toString(broadcastAddress);
        }

        public String getNetworkAddress() {
            return toString(networkAddress);
        }

        public String getNetmask() {
            return toString(netmask);
        }

        public List<String> getAllAddresses() {
            List<String> result = new ArrayList<>();

            for (long longAddress = startAddress(); longAddress <= endAddress(); longAddress++) {
                result.add(toString(longAddress));
            }

            return result;
        }

        @Override
        public String toString() {
            StringBuilder result = new StringBuilder();

            result
                .append("Network Address:\t").append(getNetworkAddress()).append("\n")
                .append("Netmask:\t\t") .append(getNetmask()).append("\n")
                .append("Broadcast Address:\t").append(getBroadcastAddress()).append("\n")
                .append("First Address:\t\t").append(toString(startAddress())).append("\n")
                .append("Last Address:\t\t").append(toString(endAddress()));

            return result.toString();
        }
        
        public boolean matchPattern(long networkPattern) {
            return (networkAddress & networkPattern) == networkPattern;
        }
        
        public String getAddress(long relativeAddress) {
            return toString(networkAddress + relativeAddress);
        }
    }
}