/*
 * Copyright (c) 2019, Oracle and/or its affiliates. All rights reserved.
 */
package com.oracle.iot.flir.utils;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.text.MessageFormat;
import java.util.*;
import java.util.Map.Entry;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.json.JSONArray;
import org.json.JSONObject;
import org.json.JSONWriter;

import com.oracle.iot.flir.model.FLIRCameraE95DeviceModelDescriptor;
import com.oracle.iot.flir.utils.http.HttpResponse;
import com.oracle.iot.flir.utils.http.HttpSecureConnection;

import oracle.iot.client.device.DirectlyConnectedDevice;

public class ProvisioningUtils {
    private static final String SHARED_SECRET = Base64.getEncoder()
            .encodeToString(UUID.randomUUID().toString().getBytes(StandardCharsets.UTF_8));
    private static final String PROVISIONING_PASSWORD = "IoTF1irR0cks1#";
    
    private static final Logger LOGGER = Logger.getLogger(ProvisioningUtils.class.getSimpleName());
    
    public static void createDeviceModel(HttpSecureConnection secureConnection, DeviceModelDescriptor dmDescriptor)
            throws IOException, GeneralSecurityException {
        String dmName = dmDescriptor.getName();
        
        LOGGER.info("Creating '" + dmName + "' device model");
        
        ByteArrayOutputStream data = new ByteArrayOutputStream();
        OutputStreamWriter writer = new OutputStreamWriter(data, StandardCharsets.UTF_8);
        JSONWriter json = new JSONWriter(writer);
        
        json
            .object()
                .key("urn").value(dmDescriptor.getURN())
                .key("name").value(dmName);
        
        Map<String, String> attributes = dmDescriptor.getAttributes();
        Map<String, String> attributesDescriptions = dmDescriptor.getAttributesDescriptions();
        
        if (attributes != null && !attributes.isEmpty()) {
            json.key("attributes").array();
            
            for (Entry<String, String> attributeEntry : attributes.entrySet()) {
                String attributeName = attributeEntry.getKey();
                
                json
                    .object()
                        .key("name").value(attributeName)
                        .key("writable").value(true)
                        .key("type").value(attributeEntry.getValue());
                
                String description = attributesDescriptions == null ? null : attributesDescriptions.get(attributeName);
                
                if (description != null) {
                    json.key("description").value(description);
                }
                
                json.endObject();
            }
            
            json.endArray();
        }
        
        json.endObject();
        
        writer.close();
        
        HttpResponse response = secureConnection.post("iot/api/v2/deviceModels", data.toByteArray());
        
        if (response.getStatus() != 201) {
            throw new IOException("Failed to create '" + dmName + "' device model");
        }
    }
    
    public static void deleteDeviceModel(HttpSecureConnection secureConnection, String urn)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Deleting device model with urm " + urn);

        HttpResponse response = secureConnection.delete("iot/api/v2/deviceModels/" + urn + "?force");
        
        if (response.getStatus() != 204) {
            throw new IOException("Cannot delete device model with urn " + urn);
        }
    }
    
    public static boolean deviceModelExists(HttpSecureConnection secureConnection, String urn)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Check if device model with URN " + urn + " exists");

        return secureConnection.get("iot/api/v2/deviceModels/" + urn).getStatus() == 200;
    }
    
    private static Set<String> getApplicationDeviceModelURNs(HttpSecureConnection secureConnection, String appID)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Getting device models of application with ID " + appID);
     
        HttpResponse response = secureConnection.get("iot/api/v2/apps/" + appID + "/deviceModels");
        
        if (response.getStatus() != 200) {
            throw new IOException("Cannot obtain device models of application with ID " + appID);
        }
        
        try {
            JSONArray items = new JSONObject(new String(response.getData(), StandardCharsets.UTF_8)).getJSONArray("items");
            LinkedHashSet<String> result = new LinkedHashSet<>();
            
            for (int i = 0; i < items.length(); i++) {
                result.add(items.getJSONObject(i).getString("urn"));
            }
            
            return result;
        } catch (Throwable t) {
            throw new IOException("Cannot obtain device models of application with ID " + appID);
        }
    }
    
    public static void ensureDeviceModelsAreAddedToApplication(HttpSecureConnection secureConnection, String appID,
            String... urns) throws IOException, GeneralSecurityException {
        if (urns == null || urns.length == 0) {
            return;
        }
        
        Set<String> deviceModelURNs = getApplicationDeviceModelURNs(secureConnection, appID);
        int existingModelsCount = deviceModelURNs.size();
        
        deviceModelURNs.addAll(Arrays.asList(urns));
        
        if (existingModelsCount < deviceModelURNs.size()) {
            LOGGER.info("Adding device models to application with ID " + appID);
            
            ByteArrayOutputStream data = new ByteArrayOutputStream();
            OutputStreamWriter writer = new OutputStreamWriter(data, StandardCharsets.UTF_8);
            JSONWriter json = new JSONWriter(writer);
            
            json
                .object()
                    .key("id").value(appID)
                    .key("deviceModelURNs").array();
            
            for (String urn : deviceModelURNs) {
                json.value(urn);
            }
            
            json.endArray().endObject();
            
            writer.close();
            
            HttpResponse response = secureConnection.patch("iot/api/v2/apps/" + appID, data.toByteArray());
          
            if (response.getStatus() != 200) {
                throw new IOException("Cannot add device models to application with ID " + appID);
            }
        }
    }
    
    public static void ensureDeviceModelsAreDeletedFromApplication(HttpSecureConnection secureConnection, String appID,
            String... urns) throws IOException, GeneralSecurityException {
        if (urns == null || urns.length == 0) {
            return;
        }
        
        Set<String> deviceModelURNs = getApplicationDeviceModelURNs(secureConnection, appID);
        int existingModelsCount = deviceModelURNs.size();
        
        deviceModelURNs.removeAll(Arrays.asList(urns));
        
        if (existingModelsCount > deviceModelURNs.size()) {
            LOGGER.info("Removing device models from application with ID " + appID);
            
            ByteArrayOutputStream data = new ByteArrayOutputStream();
            OutputStreamWriter writer = new OutputStreamWriter(data, StandardCharsets.UTF_8);
            JSONWriter json = new JSONWriter(writer);
            
            json
                .object()
                    .key("id").value(appID)
                    .key("deviceModelURNs").array();
            
            for (String urn : deviceModelURNs) {
                json.value(urn);
            }
            
            json.endArray().endObject();
            
            writer.close();
            
            HttpResponse response = secureConnection.patch("iot/api/v2/apps/" + appID, data.toByteArray());
          
            if (response.getStatus() != 200) {
                throw new IOException("Cannot remove device models from application with ID " + appID);
            }
        }
    }
    
    public static String getAMAppId(HttpSecureConnection secureConnection)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Obtaining assets monitoring app ID");
        
        HttpResponse response = secureConnection.get("iot/api/v2/apps?q={\"type\":\"assetTracking\"}");
        
        if (response.getStatus() != 200) {
            throw new IOException("Cannot obtain assets monitoring app ID");
        }
        
        try {
            return new JSONObject(new String(response.getData(), StandardCharsets.UTF_8)).getJSONArray("items")
                    .getJSONObject(0).getString("id");
        } catch (Throwable t) {
            throw new IOException("Cannot obtain production monitoring app ID");
        }
    }
    
    public static String getDeviceIDByHardwareID(HttpSecureConnection secureConnection, String hardwareID)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Getting ID of device with hardware ID " + hardwareID);
        
        HttpResponse response = secureConnection
                .get(MessageFormat.format("iot/api/v2/devices?q='{'\"hardwareId\":\"{0}\"'}'", hardwareID));
        
        if (response.getStatus() != 200) {
            throw new IOException("Cannot get ID of device with hardware ID " + hardwareID);
        }
        
        try {
            JSONObject responseObject = new JSONObject(new String(response.getData(), StandardCharsets.UTF_8));
            
            if (responseObject.getInt("count") > 0) {
                return responseObject.getJSONArray("items").getJSONObject(0).getString("id");
            } else {
                return null;
            }
        } catch (Throwable t) {
            throw new IOException("Cannot get ID of device with hardware ID " + hardwareID);
        }
    }
    
    private static boolean changeDeviceState(HttpSecureConnection secureConnection, String deviceID, String newState,
            int successStatus) throws IOException, GeneralSecurityException {
        ByteArrayOutputStream data = new ByteArrayOutputStream();
        OutputStreamWriter writer = new OutputStreamWriter(data, StandardCharsets.UTF_8);
        JSONWriter json = new JSONWriter(writer);
        
        json
            .object()
                .key("state").value(newState)
            .endObject();
            
        writer.close();
        
        HttpResponse response = secureConnection.patch("iot/api/v2/devices/" + deviceID, data.toByteArray());
        
        return response.getStatus() == successStatus;
    }
    
    public static void decommissionDevice(HttpSecureConnection secureConnection, String deviceID)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Decomissioning device with ID " + deviceID);
        
        if (!changeDeviceState(secureConnection, deviceID, "DECOMMISSIONED", 200)) {
            throw new IOException("Failed to decomissioning device with ID " + deviceID);
        }
    }
    
    public static void deleteDevice(HttpSecureConnection secureConnection, String deviceID)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Deleting device with ID " + deviceID);
        
        if (!changeDeviceState(secureConnection, deviceID, "DELETED", 204)) {
            throw new IOException("Failed to delete device with ID " + deviceID);
        }
    }
    
    public static void registerDCD(HttpSecureConnection secureConnection, DCDDescriptor deviceDescriptor)
            throws IOException, GeneralSecurityException {
        String hardwareID = deviceDescriptor.getHardwareID();
        
        LOGGER.info("Registering device " + hardwareID);
        
        ByteArrayOutputStream data = new ByteArrayOutputStream();
        OutputStreamWriter writer = new OutputStreamWriter(data, StandardCharsets.UTF_8);
        JSONWriter json = new JSONWriter(writer);
        
        json
            .object()
                .key("hardwareId").value(hardwareID)
                .key("name").value(deviceDescriptor.getName())
                .key("sharedSecret").value(SHARED_SECRET)
                .key("location")
                    .object()
                        .key("latitude").value(deviceDescriptor.getLatitude())
                        .key("longitude").value(deviceDescriptor.getLongitude())
                    .endObject();
        
        String manufacturer = deviceDescriptor.getManufacturer();
        
        if (manufacturer != null) {
            json.key("manufacturer").value(manufacturer);
        }
        
        String modelNumber = deviceDescriptor.getModelNumber();
        
        if (modelNumber != null) {
            json.key("modelNumber").value(modelNumber);
        }
        
        String serialNumber = deviceDescriptor.getSerialNumber();
        
        if (serialNumber != null) {
            json.key("serialNumber").value(serialNumber);
        }
        
        Map<String, String> metaData = deviceDescriptor.getMetaData();
        
        if (metaData != null && !metaData.isEmpty()) {
            json.key("metadata").object();
            
            for (Entry<String, String> metaDataEntry : metaData.entrySet()) {
                json.key(metaDataEntry.getKey()).value(metaDataEntry.getValue());
            }
            
            json.endObject();
        }
        
        json.endObject();
        
        writer.close();
        
        HttpResponse response = secureConnection.post("iot/api/v2/devices?expand=location,metadata",
                data.toByteArray());
        
        if (response.getStatus() != 201) {
            throw new IOException("Failed to register device " + hardwareID);
        }
        
        data = new ByteArrayOutputStream();
        writer = new OutputStreamWriter(data, StandardCharsets.UTF_8);
        json = new JSONWriter(writer);
        
        json
            .object()
                .key("id").value(hardwareID)
                .key("passphrase").value(PROVISIONING_PASSWORD)
            .endObject();
        
        writer.close();
        
        response = secureConnection.patch("iot/api/v2/provisioner/device", data.toByteArray());
        
        if (response.getStatus() != 200) {
            throw new IOException("Failed to register device " + hardwareID);
        }
        
        try (FileOutputStream out = new FileOutputStream(new File(hardwareID))) {
            out.write(response.getData());
        }
    }
    
    public static DirectlyConnectedDevice getDirectlyConnectedDevice(DCDDescriptor deviceDescriptor, String... urns) {
        LOGGER.info("Creating directly connected device");
        
        DirectlyConnectedDevice result;
        
        try {
            File provisioningFile = new File(deviceDescriptor.getHardwareID());
            
            if (!provisioningFile.canRead()) {
                return null;
            }
            
            result = new DirectlyConnectedDevice(provisioningFile.getPath(), PROVISIONING_PASSWORD);
        } catch (Throwable t) {
            LOGGER.log(Level.SEVERE, "Cannot create the directly connected device", t);

            return null;
        }
        
        if (!result.isActivated()) {
            LOGGER.info("Activating directly connected device");
            
            try {
                result.activate(urns);
            } catch (Throwable t) {
                LOGGER.log(Level.SEVERE, "Cannot activate the directly connected device", t);
                
                try {
                    result.close();
                } catch (IOException ex) {}
                
                return null;
            }
        }
        
        return result;
    }
    
    public static String getAssetTypeID(HttpSecureConnection secureConnection,
            String assetTypeName) throws IOException, GeneralSecurityException {
        LOGGER.info("Getting ID of asset type with name " + assetTypeName);
        
        HttpResponse response = secureConnection.get(MessageFormat.format(
                "iotapps/privateclientapi/v2/orgs/ORA_DEFAULT_ORG/assetTypes?q='{'\"name\":\"{0}\"'}'", assetTypeName));
        
        int status = response.getStatus();
        
        if (status == 200) {
            try {
                JSONObject responseObject = new JSONObject(new String(response.getData(), StandardCharsets.UTF_8));
                
                if (responseObject.getInt("count") > 0) {
                    return responseObject.getJSONArray("items").getJSONObject(0).getString("id");
                } else {
                    return null;
                }
            } catch (Throwable t) {
                throw new IOException("Cannot get ID of asset type with name " + assetTypeName);
            }
        } else if (status == 404) {
            return null;
        } else {
            throw new IOException("Cannot get ID of asset type with name " + assetTypeName);
        } 
    }
    
    public static String createAssetType(HttpSecureConnection secureConnection, String assetTypeName,
            DeviceModelDescriptor... dmDescrs) throws IOException, GeneralSecurityException {
        LOGGER.info("Creating asset type " + assetTypeName);
        
        ByteArrayOutputStream data = new ByteArrayOutputStream();
        OutputStreamWriter writer = new OutputStreamWriter(data, StandardCharsets.UTF_8);
        JSONWriter json = new JSONWriter(writer);

        json
            .object()
                .key("name").value(assetTypeName)
                .key("attributes")
                    .array();
        
        for (DeviceModelDescriptor dmDescr : dmDescrs) {
            for (Entry<String, String> attributeEntry : dmDescr.getAttributes().entrySet()) {
                json
                    .object()
                        .key("name").value(attributeEntry.getKey())
                        .key("static").value(false)
                        .key("type").value("SENSOR")
                        .key("properties")
                            .object()
                                .key("category").value(dmDescr.getCategories().get(attributeEntry.getKey()))
                            .endObject()
                        .key("dataType").value(attributeEntry.getValue())
                    .endObject();
            }
        }
        
        json.endArray().endObject();
            
        writer.close();
        
        HttpResponse response = secureConnection.post("iotapps/privateclientapi/v2/orgs/ORA_DEFAULT_ORG/assetTypes",
                data.toByteArray());
        
        if (response.getStatus() != 201) {
            throw new IOException("Failed to create asset type");
        }
        
        String assetTypeID;
        
        try {
            assetTypeID = new JSONObject(new String(response.getData(), StandardCharsets.UTF_8)).getString("id");
        } catch (Throwable t) {
            throw new IOException("Failed to create asset type");
        }
        
        return assetTypeID;
    }
    
    public static void deleteAssetType(HttpSecureConnection secureConnection, String assetTypeID)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Deleting asset type with ID " + assetTypeID);

        HttpResponse response = secureConnection
                .delete("iotapps/privateclientapi/v2/orgs/ORA_DEFAULT_ORG/assetTypes/" + assetTypeID);
        
        if (response.getStatus() != 204) {
            throw new IOException("Cannot delete asset type with ID " + assetTypeID);
        }
    }
    
    public static String getRuleID(HttpSecureConnection secureConnection, String ruleName)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Getting ID of rule '" + ruleName + "'");
        
        HttpResponse response = secureConnection.get(MessageFormat.format(
                "assetMonitoring/privateclientapi/v2/orgs/ORA_DEFAULT_ORG/rules?q='{'\"name\":\"{0}\"'}'", ruleName));
        
        if (response.getStatus() != 200) {
            throw new IOException("Cannot get ID of rule '" + ruleName + "'");
        }
        
        try {
            JSONObject responseObject = new JSONObject(new String(response.getData(), StandardCharsets.UTF_8));
            
            if (responseObject.getInt("count") > 0) {
                return responseObject.getJSONArray("items").getJSONObject(0).getString("id");
            } else {
                return null;
            }
        } catch (Throwable t) {
            throw new IOException("Cannot get ID of rule '" + ruleName + "'");
        }
    }
    
    public static void createTooHighTemperatureRule(Configuration config, String ruleName, String assetTypeName,
            RuleOutput ruleOutput)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Creating rule '" + ruleName + "'");
        
        ByteArrayOutputStream data = new ByteArrayOutputStream();
        OutputStreamWriter writer = new OutputStreamWriter(data, StandardCharsets.UTF_8);
        JSONWriter json = new JSONWriter(writer);

        json
            .object()
                .key("name").value(ruleName)
                .key("sources")
                    .object()
                        .key("assetType").value(assetTypeName)
                    .endObject()
                .key("predicates")
                    .array()
                        .object()
                            .key("type").value("threshold")
                            .key("config")
                                .object()
                                    .key("attributeSource").value("businessObject")
                                    .key("attributeName").value("sensor/" + FLIRCameraE95DeviceModelDescriptor.ATTRIBUTE_TEMPERATURE)
                                    .key("threshold").value(config.getMaxTemperature())
                                    .key("attributeType").value("NUMBER")
                                    .key("operator").value("ge")
                                .endObject()
                        .endObject()
                    .endArray()
                .key("predicateCombination").value("all_of");
    
        ruleOutput.toJSON(json);
        
        json.endObject();
            
        writer.close();
        
        HttpResponse response = config.getIOTConnection()
                .post("assetMonitoring/privateclientapi/v2/orgs/ORA_DEFAULT_ORG/rules", data.toByteArray());
        
        if (response.getStatus() != 201) {
            throw new IOException("Failed to create rule '" + ruleName + "'");
        }
    }
    
    public static void deleteRule(HttpSecureConnection secureConnection, String ruleID)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Deleting rule with ID " + ruleID);

        HttpResponse response = secureConnection
                .delete("assetMonitoring/privateclientapi/v2/orgs/ORA_DEFAULT_ORG/rules/" + ruleID);
        
        if (response.getStatus() != 204) {
            throw new IOException("Cannot delete rule with ID " + ruleID);
        }
    }
    
    public static void createAsset(HttpSecureConnection secureConnection, String assetName, String assetTypeName,
            Configuration config, String endpointID, DeviceModelDescriptor... dmDescrs)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Creating asset '" + assetName + "'");
        
        ByteArrayOutputStream data = new ByteArrayOutputStream();
        OutputStreamWriter writer = new OutputStreamWriter(data, StandardCharsets.UTF_8);
        JSONWriter json = new JSONWriter(writer);
        json
            .object()
                .key("name").value(assetName)
                .key("type").value(assetTypeName)
                .key("geoLocation").value(config.getCameraLatitude() + "," + config.getCameraLongitude())
                .key("attributes")
                    .array();
        
        for (DeviceModelDescriptor dmDescr : dmDescrs) {
            for (String attributeName : dmDescr.getAttributes().keySet()) {
                json
                    .object()
                        .key("name").value(attributeName)
                        .key("value")
                            .object()
                                .key("device").value("real")
                                .key("deviceModel").value(dmDescr.getURN())
                                .key("deviceId").value(endpointID)
                                .key("field").value(attributeName)
                            .endObject()
                    .endObject();
            }
        }
                        
        json.endArray().endObject();
        
        writer.close();
        
        HttpResponse response = secureConnection.post("iotapps/privateclientapi/v2/orgs/ORA_DEFAULT_ORG/assets",
                data.toByteArray());
        
        if (response.getStatus() != 201) {
            throw new IOException("Failed to create asset '" + assetName + "'");
        }
    }
    
    public static void deleteAsset(HttpSecureConnection secureConnection, String assetID)
            throws IOException, GeneralSecurityException {
        LOGGER.info("Deleting asset with ID " + assetID);

        HttpResponse response = secureConnection
                .delete("iotapps/privateclientapi/v2/orgs/ORA_DEFAULT_ORG/assets/" + assetID);
        
        if (response.getStatus() != 204) {
            throw new IOException("Cannot delete asset with ID " + assetID);
        }
    }
    
    public static String getAssetID(HttpSecureConnection secureConnection,
            String assetName) throws IOException, GeneralSecurityException {
        LOGGER.info("Getting ID of asset with name " + assetName);
        
        HttpResponse response = secureConnection.get(MessageFormat.format(
                "iotapps/privateclientapi/v2/orgs/ORA_DEFAULT_ORG/assets?q='{'\"name\":\"{0}\"'}'", assetName));
        
        int status = response.getStatus();
        
        if (status == 200) {
            try {
                JSONObject responseObject = new JSONObject(new String(response.getData(), StandardCharsets.UTF_8));
                
                if (responseObject.getInt("count") > 0) {
                    return responseObject.getJSONArray("items").getJSONObject(0).getString("id");
                } else {
                    return null;
                }
            } catch (Throwable t) {
                throw new IOException("Cannot get ID of asset with name " + assetName);
            }
        } else if (status == 404) {
            return null;
        } else {
            throw new IOException("Cannot get ID of asset with name " + assetName);
        } 
    }
    
    
    public interface DeviceDescriptor {
        String getHardwareID();
        String getManufacturer();
        String getModelNumber();
        String getSerialNumber();
        
        Map<String, String> getMetaData();
    }
    
    
    public interface DCDDescriptor extends DeviceDescriptor {
        String getName();
        String getLongitude();
        String getLatitude();
    }
    
    
    public interface DeviceModelDescriptor {
        String getURN();
        String getName();
        
        Map<String, String> getAttributes();
        Map<String, String> getAttributesDescriptions();
        Map<String, String> getCategories();
    }
    
    
    public abstract static class RuleOutput {
        protected abstract String getType();
        
        protected abstract void configKeysToJSON(JSONWriter json);
        
        public void toJSON(JSONWriter json) {
            json
                .key("actions")
                    .array()
                        .object()
                            .key("type").value(getType())
                            .key("config")
                                .object();
            
            configKeysToJSON(json);
            
            json.endObject();
            
            json.key("tags").array().endArray();
            
            json.endObject().endArray();
        }
    }
    
    
    public static class RuleIncident extends RuleOutput {
        private String summary;
        private String type;
        private String priority;
        
        public RuleIncident(String summary, String type, String priority) {
            this.summary = summary;
            this.type = type;
            this.priority = priority;
        }

        @Override
        protected String getType() {
            return "TRIGGERINCIDENT";
        }

        @Override
        protected void configKeysToJSON(JSONWriter json) {
            json
                .key("summary").value(summary)
                .key("incidentType").value(type)
                .key("priority").value(priority)
                .key("description").value("")
                .key("enableWO").value(false)
                .key("enableMCIncident").value(false)
                .key("enableSvCSIncident").value(false)
                .key("enableECSIncident").value(false);
            
        }
    }
}

