/*
 * Copyright (c) 2019, Oracle and/or its affiliates. All rights reserved.
 */
package com.oracle.iot.flir.utils.http;

import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.util.Base64;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;

public abstract class AbstractHttpConnection {
    private final String hostUrl;
    private final String basicCredentials;
    private final Integer connectTimeout;

    public AbstractHttpConnection(String hostUrl, String user, String password, Integer connectTimeout) throws GeneralSecurityException {
        this.hostUrl = hostUrl;
        this.basicCredentials = user == null || password == null ? null
                : Base64.getEncoder().encodeToString((user + ":" + password).getBytes(StandardCharsets.UTF_8));
        this.connectTimeout = connectTimeout;
    }
    
    protected abstract HttpClient createHttpClient(URL url) throws MalformedURLException;
    
    protected Integer getConnectTimeout() {
        return connectTimeout;
    }

    public HttpResponse get(String restApi) throws IOException, GeneralSecurityException {
        return invoke("GET", restApi, null);
    }
    
    public HttpResponse get(String restApi, String contentType, Map<String, String> customHeaders)
            throws IOException, GeneralSecurityException {
        return invoke("GET", restApi, null, contentType, customHeaders);
    }

    public HttpResponse post(String restApi, byte[] payload) throws IOException, GeneralSecurityException {

        return post(restApi, payload, null);
    }

    public HttpResponse post(String restApi, byte[] payload, String contentType)
            throws IOException, GeneralSecurityException {
        return invoke("POST", restApi, payload, contentType, null);
    }

    public HttpResponse put(String restApi, byte[] payload) throws IOException, GeneralSecurityException {
        return invoke("PUT", restApi, payload);
    }

    public HttpResponse delete(String restApi) throws IOException, GeneralSecurityException {
        return invoke("DELETE", restApi, null);
    }

    public HttpResponse patch(String restApi, byte[] payload) throws IOException, GeneralSecurityException {
        return invoke("PATCH", restApi, payload);
    }

    private HttpResponse invoke(String method, String restApi, byte[] payload)
            throws IOException, GeneralSecurityException {
        return invoke(method, restApi, payload, null, null);
    }

    private HttpResponse invoke(String method, String restApi, byte[] payload, String contentType,
            Map<String, String> customHeaders) throws IOException, GeneralSecurityException {
        HttpResponse response = null;
        URL url;

        try {
            final HashMap<String, String> headers = new HashMap<String, String>(4);
            url = new URL(new URL(hostUrl), restApi);
            url = new URL(new URI(url.getProtocol(), url.getUserInfo(), url.getHost(), url.getPort(), url.getPath(),
                    url.getQuery(), url.getRef()).toASCIIString());

            if (basicCredentials != null) {
                headers.put("Authorization", "Basic " + basicCredentials);
            }
            headers.put("Content-Type", contentType == null ? "application/json" : contentType);
            headers.put("Accept", "application/json, text/plain");
            
            if (customHeaders != null) {
                for (Entry<String, String> header : customHeaders.entrySet()) {
                    headers.put(header.getKey(), header.getValue());
                }
            }

            final HttpClient httpClient = createHttpClient(url);
            response = invoke(httpClient, method, payload, headers);
        } catch (URISyntaxException ex) {
            throw new IOException(ex);
        }

        return response;
    }

    protected HttpResponse invoke(HttpClient httpClient, String method, byte[] payload, Map<String, String> headers)
            throws IOException, GeneralSecurityException {

        if (method.equals("GET")) {
            return httpClient.get(headers);
        } else if (method.equals("POST")) {
            return httpClient.post(payload, headers);
        } else if (method.equals("DELETE")) {
            return httpClient.delete(headers);
        } else if (method.equals("PATCH")) {
            headers.put("X-HTTP-Method-Override", "PATCH");
            return httpClient.post(payload, headers);
        } else {
            return httpClient.put(payload, headers);
        }
    }
}

