/*
 * Copyright (c) 2019, Oracle and/or its affiliates.  All rights reserved.
 */
package com.oracle.iot.flir.utils.http;

import java.io.*;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.SocketException;
import java.net.URL;
import java.security.GeneralSecurityException;
import java.util.List;
import java.util.Map;
import java.util.zip.GZIPInputStream;
import java.util.zip.InflaterInputStream;

import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLSocketFactory;

class HttpClient {
    private final URL url;
    private static final int RESPONSE_TIMEOUT = 30000;

    private final SSLSocketFactory sslSocketFactory;
    
    private final int connectTimeout;

    HttpClient(SSLSocketFactory sslSocketFactory, URL url, Integer connectTimeout) throws MalformedURLException {
        if (url == null) {
            throw new IllegalArgumentException("url cannot be null");
        }

        this.url = sslSocketFactory == null ? url
                : new URL(null, url.toExternalForm(), new sun.net.www.protocol.https.Handler());
        this.sslSocketFactory = sslSocketFactory;
        this.connectTimeout = connectTimeout == null ? RESPONSE_TIMEOUT : connectTimeout;
    }

    private HttpResponse invokeMethod(String method, byte[] data, Map<String, String> headers)
            throws IOException, GeneralSecurityException {
        HttpURLConnection con = null;
        HttpResponse response = null;

        // If we get a SocketException, we'll retry once.
        // We have to get new HttpsURLConnection for the retry in order
        // to get a new input/output stream.
        int retries = 0;

        while (retries < 2) {
            try {
                // Create the connection object
                con = (HttpURLConnection) url.openConnection();
                
                if (con instanceof HttpsURLConnection && sslSocketFactory != null) {
                    ((HttpsURLConnection) con).setSSLSocketFactory(sslSocketFactory);
                }
                
                con.setInstanceFollowRedirects(true);

                con.setConnectTimeout(connectTimeout);
                con.setReadTimeout(RESPONSE_TIMEOUT);

                con.setRequestMethod(method);

                // Add request headers (caller-supplied first, so we overwrite any
                // collisions)
                if (headers != null) {
                    for (Map.Entry<String, String> entry : headers.entrySet()) {
                        con.setRequestProperty(entry.getKey(), entry.getValue());
                    }
                }

                if (data != null) {
                    // Send post request
                    con.setDoOutput(true);
                    con.setFixedLengthStreamingMode(data.length);
                }

                con.connect();

                if (data != null) {
                    pipe(new ByteArrayInputStream(data), con.getOutputStream());
                }

                // Check if redirection is requested
                int responseCode = con.getResponseCode();

                // Read the response
                Map<String, List<String>> responseHeaders = con.getHeaderFields();
                InputStream responseStream;
                
                if (responseCode >= 200 && responseCode < 300) {
                    try {
                        responseStream = con.getInputStream();
                    } catch (IOException ex) {
                        responseStream = con.getErrorStream();
                    }
                } else {
                    responseStream = con.getErrorStream();
                }
                
                byte[] responseData = getResponseBody(con, responseStream);

                response = new HttpResponse(responseCode, responseData, responseHeaders);
                
                break;
            } catch (SocketException se) {
                // Connection may have been reset - try again
                if (++retries == 1) {
                    continue;
                }

                throw se;
            } finally {
                con.disconnect();
            }
        }

        return response;
    }

    final HttpResponse post(byte[] data, Map<String, String> headers) throws IOException, GeneralSecurityException {
        return invokeMethod("POST", data, headers);
    }

    final HttpResponse put(byte[] data, Map<String, String> headers) throws IOException, GeneralSecurityException {
        return invokeMethod("PUT", data, headers);
    }

    final HttpResponse get(Map<String, String> headers) throws IOException, GeneralSecurityException {
        return invokeMethod("GET", null, headers);
    }

    final HttpResponse delete(Map<String, String> headers) throws IOException, GeneralSecurityException {
        return invokeMethod("DELETE", null, headers);
    }

    private void pipe(InputStream in, OutputStream out) throws IOException {
        // Setup so that if in or out is null but not both, then the one that
        // isn't null will end up being closed.
        // Otherwise any failure ends up as an IOException but the streams are
        // still closed
        try {
            if (in != null && out != null) {
                final byte[] buffer = new byte[8096];
                int length;
                while ((length = in.read(buffer)) != -1) {
                    out.write(buffer, 0, length);
                }
            }
        } finally {
            if (in != null) {
                try {
                    in.close();
                } catch (IOException ioe) {}
            }
            if (out != null) {
                try {
                    out.close();
                } catch (IOException ioe) {}
            }
        }
    }

    private byte[] getResponseBody(HttpURLConnection http, InputStream responseStream) throws IOException {
        // If this is GZIP encoded, then wrap the input stream
        final String contentEncoding = http.getContentEncoding();
        if ("gzip".equals(contentEncoding)) {
            responseStream = new GZIPInputStream(responseStream);
        } else if ("deflate".equals(contentEncoding)) {
            responseStream = new InflaterInputStream(responseStream);
        }
        ByteArrayOutputStream body = new ByteArrayOutputStream();

        pipe(responseStream, body);

        return body.toByteArray();
    }
}
